/*
Package influxql implements a parser for the InfluxDB query language.

InfluxQL is a DML and DDL language for the InfluxDB time series database.
It provides the ability to query for aggregate statistics as well as create
and configure the InfluxDB server.

See https://docs.influxdata.com/influxdb/latest/query_language/
for a reference on using InfluxQL.

*/
package influxql
